import os
import shutil
from collections.abc import ItemsView, KeysView, ValuesView
from pathlib import Path
from typing import Any, Dict, Generator, Tuple

import tomlkit
from astropy.coordinates import EarthLocation
from astropy.units import Quantity
from tomlkit.items import Table
from tomlkit.toml_document import TOMLDocument

from . import EnvVarName, get_logger
from .exceptions import NECSTConfigurationError
from .utils import ValueRange, read_file

logger = get_logger(__name__)

DefaultNECSTRoot = Path.home() / ".necst"


class _Parsers:
    observatory = str
    location = lambda x: EarthLocation(**x)  # noqa: E731
    observation_frequency = Quantity
    simulator = bool
    alert_interval_sec = float
    antenna_pid_param_az = list
    antenna_pid_param_el = list
    antenna_drive_range_az = lambda x: ValueRange(*map(Quantity, x))  # noqa: E731
    antenna_drive_range_el = lambda x: ValueRange(*map(Quantity, x))  # noqa: E731
    antenna_drive_warning_limit_az = lambda x: ValueRange(  # noqa: E731
        *map(Quantity, x)
    )
    antenna_drive_warning_limit_el = lambda x: ValueRange(  # noqa: E731
        *map(Quantity, x)
    )
    antenna_drive_critical_limit_az = lambda x: ValueRange(  # noqa: E731
        *map(Quantity, x)
    )
    antenna_drive_critical_limit_el = lambda x: ValueRange(  # noqa: E731
        *map(Quantity, x)
    )
    antenna_pointing_accuracy = Quantity
    antenna_pointing_parameter_path = Path
    antenna_scan_margin = Quantity
    antenna_max_acceleration_az = Quantity
    antenna_max_acceleration_el = Quantity
    antenna_max_speed_az = Quantity
    antenna_max_speed_el = Quantity
    antenna_speed_to_pulse_factor_az = Quantity
    antenna_speed_to_pulse_factor_el = Quantity
    antenna_command_frequency = int
    antenna_command_offset_sec = float
    ros_service_timeout_sec = float
    ros_communication_deadline_sec = float
    ros_logging_interval_sec = float
    ros_topic_scan_interval_sec = float


class Configuration:

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        self.__config = None
        self.reload()

    def __getitem__(self, key: str) -> Any:
        ret = self.__config.__getitem__(key)
        return self._dotnecst + f"/{ret}" if isinstance(ret, Path) else ret

    def __getattr__(self, key: str) -> Any:
        ret = getattr(self.__config, key)
        return self._dotnecst + f"/{ret}" if isinstance(ret, Path) else ret

    @property
    def _dotnecst(self) -> Path:
        return self.__config_path.rsplit("/", 1)[0]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(file='{self.__config_path}')"

    def __str__(self) -> str:
        return str(self.__config)

    def reload(self) -> None:
        params = self.__find_config()
        new_config = _Cfg(self, params)
        # Attribute validation
        attr_cross_section = set(dir(self.__class__)) & set(new_config.keys())
        dupl_attrs = list(filter(lambda x: not x.startswith("_"), attr_cross_section))
        if dupl_attrs:
            raise NECSTConfigurationError(
                f"Parameters {dupl_attrs} cannot be set; reserved name."
            )
        # Assign after validation
        self.__config = new_config

    def __find_config(self) -> TOMLDocument:
        root_candidates = [DefaultNECSTRoot]
        if EnvVarName.necst_root in os.environ:
            root_candidates.insert(0, os.environ[EnvVarName.necst_root])
        config_file_name = "config.toml"
        candidates = map(lambda x: str(x) + f"/{config_file_name}", root_candidates)

        for path in candidates:
            try:
                contents = read_file(path, localonly=False)
                logger.info(f"Imported configuration file '{path}'")
                self.__config_path = path
                return tomlkit.parse(contents)
            except FileNotFoundError:
                logger.debug(f"Config file not found at {path!r}")

        logger.error(
            "Config file not found, using the default parameters. "
            "To create the file with default parameters, run `neclib.configure()`."
        )
        self.__config_path = str(Path(__file__).parent / "src" / "config.toml")
        return tomlkit.parse(read_file(self.__config_path, localonly=True))

    @classmethod
    def configure(cls) -> None:
        """Create config file under ``$HOME/.necst``"""
        DefaultNECSTRoot.mkdir(exist_ok=True)
        for filename in ["config.toml", "pointing_param.toml"]:
            _target_path = DefaultNECSTRoot / filename
            if _target_path.exists():
                logger.error(f"'{_target_path}' already exists, skipping...")
                continue
            shutil.copyfile(Path(__file__).parent / "src" / filename, _target_path)
        cls().reload()


class _Cfg:
    __slots__ = ("_raw_config", "_config_manager", "_prefix", "_config")

    def __init__(
        self,
        config_manager: Configuration,
        config: TOMLDocument,
        prefix: str = "",
        parsed: Dict[str, Any] = None,
    ) -> None:
        self._raw_config = config
        self._config_manager = config_manager
        self._prefix = prefix
        if parsed is None:
            parsed = dict(self.__flatten(config))
        self._config = {}
        for key, value in parsed.items():
            if not self.__norm(key).startswith(self.__norm(self._prefix)):
                continue
            self._config[key] = value

        config_keys = [self.__norm(key) for key in self._config.keys()]
        duplicated_key = [key for key in set(config_keys) if config_keys.count(key) > 1]
        if len(duplicated_key) > 0:
            raise NECSTConfigurationError(
                "Duplicated keys are not allowed in the configuration file: "
                f"{duplicated_key}"
            )

    def __flatten(
        self, data: Dict[str, Any], prefix: str = "", raw: bool = False
    ) -> Generator[Tuple[str, Any], None, None]:
        for key, value in data.items():
            if isinstance(value, Table):
                yield from self.__flatten(value, prefix + key + ".", raw=raw)
            else:
                if raw:
                    yield prefix + key, value
                    continue
                value = value.unwrap() if hasattr(value, "unwrap") else value
                parser = getattr(
                    _Parsers, (prefix + key).replace(".", "_"), lambda x: x
                )
                if (
                    not (prefix + key)
                    .replace(".", "_")
                    .startswith(self._prefix.replace(".", "_"))
                ):
                    continue
                parsed = parser(value)
                parsed = (
                    self._config_manager._dotnecst + f"/{parsed}"
                    if isinstance(parsed, Path)
                    else parsed
                )
                yield prefix + key, parsed

    @property
    def _raw_flat(self) -> Generator[Tuple[str, Any], None, None]:
        yield from self.__flatten(self._raw_config, raw=True)

    def __str__(self) -> str:
        width = max(len(p) for p in self._config.keys()) - len(self._prefix) + 2

        def _prettify(key: str) -> str:
            key, value = key[len(self._prefix) :], self._config.get(key)
            return f"    {key:{width}s}{value!s}    ({type(value).__name__})"

        _parameters = "\n".join(map(_prettify, self._config.keys()))
        return f"NECST configuration (prefix='{self._prefix}')\n{_parameters}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(prefix={self._prefix!r})"

    def __norm(self, key: str) -> str:
        return key.replace(".", "_").lower()

    def __getitem__(self, key: str) -> Any:
        full_key = self._prefix + key
        if full_key in self._config:
            return self._config[full_key]
        if self.__norm(full_key) in map(lambda x: self.__norm(x), self._config):
            for k, v in self._config.items():
                if self.__norm(k) == self.__norm(full_key):
                    return v

        same_prefix = [k for k in self._config if k.startswith(full_key + ".")]
        full_key += "." if len(same_prefix) > 0 else "_"
        return _Cfg(self._config_manager, self._raw_config, full_key)

    def __getattr__(self, key: str) -> Any:
        return self[key]

    def keys(self, full: bool = False) -> KeysView:
        if full:
            return self._config.keys()
        return {k[len(self._prefix) :]: None for k in self._config.keys()}.keys()

    def values(self, full: bool = False) -> ValuesView:
        return self._config.values()

    def items(self, full: bool = False) -> ItemsView:
        if full:
            return self._config.items()
        return {k[len(self._prefix) :]: v for k, v in self._config.items()}.items()

    def __gt__(self, other: Any) -> bool:
        if other is None:
            return True
        if not isinstance(other, (_Cfg, Configuration)):
            return NotImplemented
        if set(self.keys(full=False)) <= set(other.keys(full=False)):
            return False
        lazy_eq = (getattr(self, k) == getattr(other, k) for k in other.keys())
        full_eq = (getattr(self, k) == getattr(other, k) for k in other.keys(full=True))
        return True if all(lazy_eq) or all(full_eq) else False

    def __lt__(self, other: Any) -> bool:
        if other is None:
            return False
        if not isinstance(other, (_Cfg, Configuration)):
            return NotImplemented
        if set(self.keys(full=False)) >= set(other.keys(full=False)):
            return False
        lazy_eq = (getattr(self, k) == getattr(other, k) for k in self.keys())
        full_eq = (getattr(self, k) == getattr(other, k) for k in self.keys(full=True))
        return True if all(lazy_eq) or all(full_eq) else False

    def __eq__(self, other: Any) -> bool:
        if other is None:
            return False
        if not isinstance(other, (_Cfg, Configuration)):
            return NotImplemented
        if set(self.keys(full=False)) != set(other.keys(full=False)):
            return False
        lazy_eq = (getattr(self, k) == getattr(other, k) for k in self.keys())
        full_eq = (getattr(self, k) == getattr(other, k) for k in self.keys(full=True))
        return True if all(lazy_eq) or all(full_eq) else False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __ge__(self, other: Any) -> bool:
        return self.__eq__(other) or self.__gt__(other)

    def __le__(self, other: Any) -> bool:
        return self.__eq__(other) or self.__lt__(other)

    def __add__(self, other: Any) -> "_Cfg":
        if other is None:
            return self
        if not isinstance(other, (_Cfg, Configuration)):
            return NotImplemented

        if self._prefix == other._prefix:
            new = TOMLDocument()
            for k, v in self._raw_flat:
                if self.__norm(k).startswith(self.__norm(self._prefix)):
                    new[k] = v
            for k, v in other._raw_flat:
                if self.__norm(k).startswith(self.__norm(other._prefix)):
                    if (k in new.keys()) and (new[k] != v):
                        raise ValueError(
                            "Cannot merge configurations with conflicting values"
                        )
                    new[k] = v
            return _Cfg(self._config_manager, new, self._prefix, self._config)
        else:
            new = TOMLDocument()
            self_prefix_len = len(self._prefix)
            other_prefix_len = len(other._prefix)
            for k, v in self._raw_flat:
                if self.__norm(k).startswith(self.__norm(self._prefix)):
                    new[k[self_prefix_len:]] = v
            for k, v in other._raw_flat:
                if self.__norm(k).startswith(self.__norm(other._prefix)):
                    if (k in new.keys()) and (new[k] != v):
                        raise ValueError(
                            "Cannot merge configurations with conflicting values"
                        )
                    new[k[other_prefix_len:]] = v
            return _Cfg(self._config_manager, new, "")

    __radd__ = __add__


config = Configuration()
configure = Configuration.configure



import time
from typing import Optional, Union

import astropy.units as u
import ogameasure

from ...units import dBm
from ...utils import busy, skip_on_simulator
from .signal_generator_base import SignalGenerator


class FSW0010(SignalGenerator):

    Manufacturer: str = "PhaseMatrix"
    Model = "FSW0010"

    Identifier = "host"

    @skip_on_simulator
    def __init__(self, **kwargs) -> None:
        com = ogameasure.ethernet(self.Config.host, self.Config.port)
        self.sg = ogameasure.Phasematrix.FSW0010(com)
        self.sg.use_external_reference_source()

    def set_freq(self, GHz: Union[int, float]) -> None:
        with busy(self, "busy"):
            self.sg.freq_set(GHz)
            time.sleep(0.1)

    def set_power(self, dBm: Union[int, float]) -> None:
        with busy(self, "busy"):
            self.sg.power_set(dBm)
            time.sleep(0.1)

    def get_freq(self) -> u.Quantity:
        with busy(self, "busy"):
            f = self.sg.freq_query()
            time.sleep(0.1)
            return f * u.Hz

    def get_power(self) -> u.Quantity:
        with busy(self, "busy"):
            f = self.sg.power_query()
            time.sleep(0.1)
            return f * dBm

    def start_output(self) -> None:
        with busy(self, "busy"):
            self.sg.output_on()
            time.sleep(0.1)

    def stop_output(self) -> None:
        with busy(self, "busy"):
            self.sg.output_off()
            time.sleep(0.1)

    def get_output_status(self) -> Optional[bool]:
        with busy(self, "busy"):
            f = self.sg.output_query()
            time.sleep(0.1)
            if f == 1:
                return True
            elif f == 0:
                return False
            else:
                return None

    def finalize(self) -> None:
        self.stop_output()
        try:
            self.sg.com.close()
        except AttributeError:
            pass



from .convert import *  # noqa: F401, F403
from .frame import *  # noqa: F401, F403
from .observer import *  # noqa: F401, F403
from .optimize import *  # noqa: F401, F403
from .path_finder import *  # noqa: F401, F403



from abc import ABC, abstractmethod
from collections import UserDict
from collections.abc import KeysView
from functools import partial
from typing import Any, ClassVar, Dict, List, Optional, Union, final

from ..configuration import Configuration, _Cfg, config


def get_device_list() -> Dict[str, "DeviceBase"]:
    # Unlike __mro__, __subclasses__ returns immediate subclasses only.
    device_kinds = DeviceBase.__subclasses__()
    table = {}
    for v in device_kinds:
        for impl in v.__subclasses__():
            table[impl.Model.lower()] = impl
    return table


def find_config(
    key: str, identifier: Optional[str] = None
) -> Union[Configuration, _Cfg]:
    if identifier is None:
        return config[key]
    devices = {k[:-2]: v for k, v in config.items() if k.endswith("._")}
    same_model = {k: config[k] for k in devices if config[k]._ == config[key]._}
    same_machine = {
        k: v
        for k, v in same_model.items()
        if config[k][identifier] == config[key][identifier]
    }
    cfg = None
    for _cfg in same_machine.values():
        cfg += _cfg
    return cfg


class DeviceMapping(UserDict):
    def __getattr__(self, key: str) -> Dict[str, Any]:
        targets = {k: getattr(v, key) for k, v in self.items()}
        if all(callable(t) for t in targets.values()):

            def func(*args, **kwargs) -> List[Any]:
                if (
                    ("id" in kwargs)
                    and isinstance(kwargs["id"], str)
                    and (len(kwargs["id"].split(".")) == 2)
                ):
                    device_id, ch_id = kwargs["id"].split(".")
                    kwargs["id"] = ch_id
                    return targets[device_id](*args, **kwargs)
                return {k: t(*args, **kwargs) for k, t in targets.items()}

            return func
        else:
            return targets

    def __call__(self, *args, **kwargs) -> "DeviceMapping[str, Any]":
        if (
            ("id" in kwargs)
            and isinstance(kwargs["id"], str)
            and (len(kwargs["id"].split(".")) == 2)
        ):
            device_id, ch_id = kwargs["id"].split(".")
            kwargs["id"] = ch_id
            return self[device_id](*args, **kwargs)
        return DeviceMapping((k, v(*args, **kwargs)) for k, v in self.items())

    def __repr__(self) -> str:
        items = ", ".join([f"{k}={v!r}" for k, v in self.items()])
        return f"DeviceMapping({items})"

    def __str__(self) -> str:
        items = ", ".join([f"{k}={v!s}" for k, v in self.items()])
        return f"DeviceMapping({items})"

    def keys(self) -> KeysView:
        _all = {}
        for devname, dev in super().items():
            channels = dev.keys()
            if len(channels) > 0:
                for ch in channels:
                    _all[f"{devname}.{ch}"] = None
            else:
                _all[devname] = None
        return _all.keys()


class ConfigManager:
    def __set_name__(self, owner, name):
        self.name = "_" + name

    def __get__(self, instance, owner):
        return getattr(instance, self.name)

    def __set__(self, instance, value):
        setattr(instance, self.name, value)


class DeviceBase(ABC):

    _instances: Dict[str, "DeviceBase"] = {}
    Identifier: Optional[str] = None
    Config: ConfigManager()

    Manufacturer: ClassVar[str] = ""
    Model: ClassVar[str]

    @final
    def __new__(
        cls,
        *,
        name: Optional[str] = None,
        model: Optional[Union[str, Dict[str, str]]] = None,
    ):
        if cls is not DeviceBase:
            if name is None:
                raise ValueError("Name for this device must be specified")
            cfg = find_config(name, cls.Identifier)
            key = f"{cls.Manufacturer}.{cls.Model}.{getattr(cfg, cls.Identifier, None)}"
            if key in cls._instances:
                cls._instances[key].__init__ = lambda *args, **kwargs: None
                return cls._instances[key]

            cls._instances[key] = super().__new__(cls)
            cls._instances[key].Config = cfg
            return cls._instances[key]

        impl = get_device_list()

        if isinstance(model, dict):
            name = {k: f"{name}.{k}" for k in model}
            parsed = {
                k: partial(impl[v.lower()], name=name[k]) for k, v in model.items()
            }
            return DeviceMapping(parsed)

        parsed = partial(impl[model.lower()], name=name)
        return parsed

    @final
    def __getitem__(self, key: str) -> "DeviceBase":
        # TODO: This should return collection of `partial` methods or its effective
        # equivalents, with `id` arguments set like `DeviceMapping` does.
        raise NotImplementedError

    @final
    def __repr__(self) -> str:
        model = f"model={self.Model!r}"
        manufacturer = f"manufacturer={self.Manufacturer!r}"
        ident = ""
        if self.Identifier is not None:
            ident = f"{self.Identifier}={getattr(self.Config, self.Identifier, None)!r}"
        return f"{self.__class__.__name__}({model}, {manufacturer}, {ident})"

    @final
    def __str__(self) -> str:
        return self.__repr__()

    @abstractmethod
    def finalize(self) -> None:
        ...

    @final
    def keys(self) -> KeysView:
        return getattr(self.Config, "channel", {}).keys()



"""Pointing error correction.

The telescope pointing is critically important in performing accurate and uniform
observation. This module employs following pointing model:

"""

import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from typing import Optional, Tuple, final

import astropy.units as u
import numpy as np
import tomlkit
from astropy.coordinates import Angle
from tomlkit.exceptions import UnexpectedCharError

from .. import utils
from ..typing import QuantityValue, UnitType


@dataclass
class PointingError(ABC):

    model: str
    file: Optional[os.PathLike] = None

    def __new__(cls, model: str, **kwargs):
        model = utils.toCamelCase(model).lower()
        if model == cls.__name__.lower():
            return super().__new__(cls)

        subcls = {v.__name__.lower(): v for v in cls.__subclasses__()}
        if model in subcls:
            return dataclass(subcls[model])(model, **kwargs)
        raise TypeError(
            f"Unknown pointing model: {model!r}; supported: {list(subcls.keys())}"
        )

    @final
    @classmethod
    def from_file(cls, filename: os.PathLike, model: Optional[str] = None, **kwargs):
        contents = utils.read_file(filename)
        kwargs.update({"file": filename})
        try:
            parsed = tomlkit.parse(contents)
            model = model or parsed.get("model", None)
            parsed = parsed.get("pointing_params", parsed)
            quantities = {}
            for k, v in parsed.items():
                if v == {}:  # Empty value.
                    quantities[k] = None
                    continue
                try:
                    quantities[k] = Angle(v)
                except (u.UnitsError, u.UnitTypeError, ValueError):
                    quantities[k] = u.Quantity(v)
                    # Boolean values can be parsed as quantities, but we currently don't
                    # expect such values to be in the pointing model parameters.
                except TypeError:
                    quantities[k] = v
        except UnexpectedCharError:
            parsed_to = cls(model).__class__
            units = {
                f.name: getattr(f.default, "unit", u.dimensionless_unscaled)
                for f in fields(parsed_to)[2:]  # Not 'model' or 'file'
            }
            normalized_contents = filter(lambda x: x.strip(), contents.split("\n"))
            quantities = {
                k: float(v) * unit
                for (k, unit), v in zip(units.items(), normalized_contents)
            }
        return cls(model, **quantities, **kwargs)

    def __getitem__(self, key):
        if key in (f.name for f in fields(self)):
            return getattr(self, key)

    @abstractmethod
    def fit(self, *args, **kwargs):
        """Fit the model to the data."""
        ...

    @abstractmethod
    def offset(self, az: u.Quantity, el: u.Quantity) -> Tuple[u.Quantity, u.Quantity]:
        ...

    def apparent2refracted(
        self, az: QuantityValue, el: QuantityValue, unit: Optional[UnitType] = None
    ) -> Tuple[u.Quantity, u.Quantity]:
        az, el = utils.get_quantity(az, el, unit=unit)
        dAz, dEl = self.offset(az, el)
        return az + dAz, el + dEl

    def refracted2apparent(
        self, az: QuantityValue, el: QuantityValue, unit: Optional[UnitType] = None
    ) -> Tuple[u.Quantity, u.Quantity]:
        az, el = utils.get_quantity(az, el, unit=unit)
        dAz, dEl = self.offset(az, el)
        return az - dAz, el - dEl


class NANTEN2(PointingError):
    r"""Pointing model used in NANTEN2.

    .. math::

        \Delta x =& \chi_{Az} \sin ( \omega_{Az} - Az ) \sin ( El ) \\
        &+ \epsilon \sin ( El ) \\
        &+ \chi_{2, Az} \sin ( 2 ( \omega_{2, Az} - Az ) ) \sin ( El ) \\
        &+ \mathrm{d}Az \cos ( El ) \\
        &+ \mathrm{d}e \\
        &+ \mathrm{cor}_v \cos ( El + \mathrm{cor}_p ) \\
        &+ \mathrm{d}e_\mathrm{radio} \\
        \Delta Az =& \Delta x / \cos ( El ) \\
        \Delta y =& - \chi_{El} \cos ( \omega_{El} - Az ) \\
        &- \chi_{2, El} \cos ( 2 ( \omega_{2, El} - Az ) ) \\
        &+ g_1 \cos ( El ) + g_2 \sin ( El ) \\
        &+ \mathrm{d}el \\
        &+ g_{ 1,\mathrm{radio} } \cos ( El ) + g_{ 2,\mathrm{radio} } \sin ( El ) \\
        &- \mathrm{cor}_v \sin ( El + \mathrm{cor}_p ) \\
        &+ \mathrm{d}el_\mathrm{radio} \\
        \Delta El =& \Delta y

    """

    dAz: Angle = 0 << u.arcsec
    """Azimuth (not X) offset of encoder reading."""
    de: Angle = 0 << u.arcsec
    """X collimation error."""
    chi_Az: Angle = 0 << u.arcsec
    """Magnitude of tilt of azimuth axis."""
    omega_Az: Angle = 0 << u.deg
    """Phase (azimuth direction) of tilt of azimuth axis."""
    eps: Angle = 0 << u.arcsec
    """Skew angle (lack of orthogonality) between azimuth and elevation axes."""
    chi2_Az: Angle = 0 << u.arcsec
    """Same as chi, but the period is 180deg in azimuth axis (harmonic component)."""
    omega2_Az: Angle = 0 << u.deg
    """Same as omega, but the period is 180deg in azimuth (harmonic component)."""
    chi_El: Angle = 0 << u.arcsec
    """Amplitude of tilt of azimuth axis (same as chi_Az)."""
    omega_El: Angle = 0 << u.deg
    """Phase (azimuth direction) of tilt of azimuth axis (same as omega_Az)."""
    chi2_El: Angle = 0 << u.arcsec
    """Same as chi, but the period is 180deg in azimuth axis (harmonic component, same
    as chi2_Az)."""
    omega2_El: Angle = 0 << u.deg
    """Same as omega, but the period is 180deg in azimuth (harmonic component, same as
    omega2_Az)."""
    g: float = 0
    """First order gravitational deflection coefficient."""
    gg: float = 0
    """Second order gravitational deflection coefficient."""
    ggg: float = 0
    """Third order gravitational deflection coefficient."""
    gggg: float = 0
    """Fourth order gravitational deflection coefficient."""
    dEl: Angle = 0 << u.arcsec
    """Elevation offset of encoder reading."""
    de_radio: Angle = 0 << u.arcsec
    """Constant X (not azimuth) offset between optical and radio beam."""
    del_radio: Angle = 0 << u.arcsec
    """Constant elevation offset between optical and radio beam."""
    cor_v: Angle = 0 << u.arcsec
    """Amplitude of collimation error."""
    cor_p: Angle = 0 << u.deg
    """Phase of collimation error, negative of elevation where the elevation component
    of collimation error is zero."""
    g_radio: float = 0
    """First order gravitational deflection coefficient."""
    gg_radio: float = 0
    """Second order gravitational deflection coefficient."""
    ggg_radio: float = 0
    """Third order gravitational deflection coefficient."""
    gggg_radio: float = 0
    """Fourth order gravitational deflection coefficient."""

    def offset(self, az: u.Quantity, el: u.Quantity) -> Tuple[u.Quantity, u.Quantity]:
        gravitational_term = np.polynomial.Polynomial(
            [0, self.g, self.gg, self.ggg, self.gggg]
        )
        radio_gravitational_term = np.polynomial.Polynomial(
            [0, self.g_radio, self.gg_radio, self.ggg_radio, self.gggg_radio]
        )

        dx = (
            self.chi_Az * np.sin(self.omega_Az - az) * np.sin(el)
            + self.eps * np.sin(el)
            + self.chi2_Az * np.sin(2 * (self.omega2_Az - az)) * np.sin(el)
            + self.dAz * np.cos(el)
            + self.de
            + self.cor_v * np.cos(el + self.cor_p)
            + self.de_radio
        )  # NOTE: 3rd term: chi2 sin sin, revised from chi2 cos sin
        dAz = dx / np.cos(el)
        dEl = (
            -1 * self.chi_El * np.cos(self.omega_El - az)
            - self.chi2_El * np.cos(2 * (self.omega2_El - az))
            + gravitational_term(el.to_value("deg")) * u.arcsec
            + self.dEl
            + radio_gravitational_term(el.to_value("deg")) * u.arcsec
            - self.cor_v * np.sin(el + self.cor_p)
            + self.del_radio
        )  # NOTE: 2nd term: chi2 cos, revised from chi2 sin
        # TODO: Review dimension of gravitational terms
        return dAz, dEl

    def fit(self, *args, **kwargs):
        raise NotImplementedError("Fitting is not implemented for this model.")


class OMU1P85M(PointingError):
    r"""Pointing model used in OMU1P85M.

    .. math::

        \Delta x =& a_1 \sin ( El ) \\
        &+ a_2 \\
        &+ a_3 \cos ( El ) \\
        &+ b_1 \sin ( Az ) \sin ( El ) \\
        &- b_2 \cos ( Az ) \sin ( El ) \\
        &+ c_1 * \sin ( Az - El ) \\
        &+ c_2 * \cos ( Az - El ) \\
        &+ d_1 \\
        &+ e_1 \cos ( El ) \\
        &- e_2 \sin ( El ) \\
        \Delta Az =& \Delta x / \cos ( El ) \\
        \Delta y =& b_1 \cos ( Az ) \\
        &+ b_2 \sin ( Az ) \\
        &+ b_3 \\
        &+ g_1 El \\
        &+ c_1 \cos ( Az - El ) \\
        &- c_2 \sin ( Az - El ) \\
        &+ d_2 \\
        &+ e_1 \sin ( El ) \\
        &+ e_2 \cos ( El ) \\
        \Delta El =& \Delta y

    """

    a1: Angle = 0 << u.deg
    a2: Angle = 0 << u.deg
    a3: Angle = 0 << u.deg
    b1: Angle = 0 << u.deg
    b2: Angle = 0 << u.deg
    b3: Angle = 0 << u.deg
    g1: float = 0
    c1: Angle = 0 << u.deg
    c2: Angle = 0 << u.deg
    d1: Angle = 0 << u.deg
    d2: Angle = 0 << u.deg
    e1: Angle = 0 << u.deg
    e2: Angle = 0 << u.deg

    def offset(self, az: u.Quantity, el: u.Quantity) -> Tuple[u.Quantity, u.Quantity]:
        dx = (
            self.a1 * np.sin(el)
            + self.a2
            + self.a3 * np.cos(el)
            + self.b1 * np.sin(az) * np.sin(el)
            - self.b2 * np.cos(az) * np.sin(el)
            + self.c1 * np.sin(az - el)
            + self.c2 * np.cos(az - el)
            + self.d1
            + self.e1 * np.cos(el)
            - self.e2 * np.sin(el)
        )
        dAz = dx / np.cos(el)
        dy = (
            self.b1 * np.cos(az)
            + self.b2 * np.sin(az)
            + self.b3
            + self.g1 * el
            + self.c1 * np.cos(az - el)
            - self.c2 * np.sin(az - el)
            + self.d2
            + self.e1 * np.sin(el)
            + self.e2 * np.cos(el)
        )
        dEl = dy
        return dAz, dEl

    def fit(self, *args, **kwargs):
        raise NotImplementedError("Fitting is not implemented for this model.")



import astropy.units as u
import ogameasure

from ... import utils
from .attenuator_base import Attenuator


class RHIO10(Attenuator):

    Manufacturer: str = "SENA"
    Model: str = "RHIO10"

    Identifier = "host"

    @utils.skip_on_simulator
    def __init__(self, **kwargs) -> None:
        com = ogameasure.ethernet(self.Config.host, self.Config.port)
        self.io = ogameasure.SENA.adios(com)

    def get_loss(self, id: str) -> u.Quantity:
        ch = self.Config.channel[id]
        if ch == 1:
            return self.io.get_att1() << u.dB
        elif ch == 2:
            return self.io.get_att2() << u.dB
        raise ValueError(f"Invalid channel: {ch}")

    def set_loss(self, dB: int, id: str) -> None:
        ch = self.Config.channel[id]
        self.io._set_att(ch, int(dB))

    def finalize(self) -> None:
        self.io.com.close()
